/* Copyright 2017 Leandro Augusto Frata Fernandes.
 * Distributed under the GNU General Public License, Version 3.
 *
 * Last update: October 30th, 2017.
 *
 * This implementation is part of the material accompanying the book:
 *
 *    lgebra Geomtrica e Aplicaes
 *    Fernandes, L. A. F.; Lavor, C.; Oliveira, M. M.
 *
 *    Sociedade Brasileira de Matemtica Aplicada e Computacional (SBMAC)
 *    Notas em Matemtica Aplicada (NoMA), Volume 85, 2017
 *
 *    ISSN: 2175-3385 | e-ISSN: 2236-5915
 *    ISBN: 978-85-8215-081-8 | e-ISBN: 978-85-8215-080-1
 *
 *    http://www.ic.uff.br/algebrageometrica
 */

#ifndef GA_HPP
#define GA_HPP

#include <map>
#include <algorithm>
#include <cstdint>
#include <functional>
#include <ostream>
#include <stdexcept>

#ifndef GA_DEFAULT_TOLERANCE
	#define GA_DEFAULT_TOLERANCE 1.0e-8
#else
	static_assert(GA_DEFAULT_TOLERANCE >= 0, "GA_DEFAULT_TOLERANCE must be a non-negative value.")
#endif // GA_DEFAULT_TOLERANCE

namespace ga {

	// Some useful types.

	typedef int32_t  entry_t;

	typedef uint64_t mask_t;
	typedef double   coefficient_t;

	typedef uint32_t dims_t;
	typedef dims_t   index_t;

	typedef int32_t  grade_t;

	// Some pre-declarations.

	class orthogonal_metric;
	class multivector;

	namespace detail {

		multivector _graded_gp(const multivector&, const multivector&, const orthogonal_metric&, std::function<bool(grade_t, grade_t, grade_t)>);
		multivector _graded_uminus(const multivector&, std::function<bool(grade_t)>);
		multivector _merge(const multivector&, const multivector&, std::function<coefficient_t(coefficient_t, coefficient_t)>);
		mask_t _used_basis_vectors(const multivector&);

	}

	// The superclass for metric matrix definition.
	class metric {
	public:

		typedef entry_t entry_type;

		// Return a matrix entry.
		virtual entry_type matrix_entry(index_t row, index_t col) const = 0;
	};

	// The superclass for orhogonal matrix definition.
	class orthogonal_metric : public metric {
	public:

		typedef entry_type entry_type;

		// Return a matrix entry
		constexpr entry_type matrix_entry(index_t row, index_t col) const override {
			return row == col ? diagonal_entry(row) : static_cast<entry_type>(0);
		}

		// Return a matrix entry on the main diagonal.
		virtual entry_type diagonal_entry(index_t index) const = 0;

		// Return the metric factor computed for the given basis blade.
		virtual entry_type metric_factor(mask_t mask) const = 0;
	};

	// Orhogonal metric matrix having a given p,q signature.
	template<dims_t P, dims_t Q>
	class signed_metric : public orthogonal_metric {
	public:

		typedef entry_t entry_type;

		static const dims_t p = P;
		static const dims_t q = Q;

		// Return a matrix entry on the main diagonal.
		constexpr entry_type diagonal_entry(index_t index) const override {
			if (index < 1) {
				throw std::invalid_argument("The first basis vector index is 1.");
			}
			return static_cast<entry_type>(index <= p ? +1 : (index <= p + q ? -1 : 0));
		}

		// Return the metric factor computed for the given basis blade.
		inline entry_type metric_factor(mask_t mask) const override {
			index_t index = 0;
			entry_type result = static_cast<entry_type>(1);
			while (mask != mask_t(0)) {
				if ((mask & mask_t(1)) != mask_t(0)) {
					result *= diagonal_entry(index);
				}
				mask >>= 1;
				index++;
			}
			return result;
		}
	};

	// Euclidean metric matrix.
	class euclidean_metric : public orthogonal_metric {
	public:

		typedef entry_t entry_type;

		// Return a matrix entry on the main diagonal.
		constexpr entry_type diagonal_entry(index_t) const override {
			return static_cast<entry_type>(1);
		}

		// Return the metric factor computed for the given basis blade.
		constexpr entry_type metric_factor(mask_t) const override {
			return static_cast<entry_type>(1);
		}
	};

	// Multivector defines a sparce multivector structure.
	class multivector {
	public:

		typedef mask_t mask_type;
		typedef coefficient_t coefficient_type;

		typedef typename std::map<mask_type, coefficient_type>::const_iterator const_iterator;
		typedef typename std::map<mask_type, coefficient_type>::const_reverse_iterator const_reverse_iterator;

		// Default constructor.
		inline multivector() :
			components_() {}

		// Copy constructor.
		inline multivector(const multivector& other) :
			components_(other.components_) {}

		// Initialization constructor.
		inline multivector(coefficient_t arg) :
			components_({ {mask_t(0), arg} }) {}

		// Return an iterator on the begin of a set like object providing a view on multivector's components.
		inline const_iterator begin() const {
			return components_.begin();
		}

		// Return an iterator on the end of a set like object providing a view on multivector's components.
		inline const_iterator end() const {
			return components_.end();
		}

		// Return a reverse iterator on the begin of a set like object providing a view on multivector's components.
		inline const_reverse_iterator rbegin() const {
			return components_.rbegin();
		}

		// Return a reverse iterator on the end of a set like object providing a view on multivector's components.
		inline const_reverse_iterator rend() const {
			return components_.rend();
		}

	private:

		explicit multivector(std::map<mask_type, coefficient_type>&& components) :
			components_(components) {}

		std::map<mask_type, coefficient_type> components_;

		friend coefficient_t native(const multivector&, coefficient_t);
		friend multivector scalar(coefficient_t);

		friend multivector e(index_t);

		friend multivector take_grade(const multivector&, grade_t);

		friend multivector detail::_graded_gp(const multivector&, const multivector&, const orthogonal_metric&, std::function<bool(grade_t, grade_t, grade_t)>);
		friend multivector detail::_graded_uminus(const multivector&, std::function<bool(grade_t)>);
		friend multivector detail::_merge(const multivector&, const multivector&, std::function<coefficient_t(coefficient_t, coefficient_t)>);

		friend std::ostream& operator<< (std::ostream&, const multivector&);
	};

	// Default tolerance value for floating-point comparison.
	constexpr coefficient_t default_tolerance() {
		return static_cast<coefficient_t>(GA_DEFAULT_TOLERANCE);
	}

	// Return whether the given argument is a scalar value.
	inline bool is_scalar(const multivector& arg, coefficient_t tol = default_tolerance()) {
		return std::all_of(arg.begin(), arg.end(), [&] (const std::pair<mask_t, coefficient_t>& comp) {
				return (comp.first == mask_t(0) || std::abs(comp.second) <= tol);
			});
	}

	// Return whether the given argument is zero.
	inline bool is_zero(const multivector& arg, coefficient_t tol = default_tolerance()) {
		return std::all_of(arg.begin(), arg.end(), [&] (const std::pair<mask_t, coefficient_t>& comp) {
				return std::abs(comp.second) <= tol;
			});
	}

	// Scalar multivector to number conversion.
	inline coefficient_t native(const multivector& arg, coefficient_t tol = default_tolerance()) {
		if (!is_scalar(arg, tol)) {
			throw std::invalid_argument("The multivector must be a scalar value.");
		}
		auto itr = arg.components_.find(0);
		return (itr != arg.components_.end()) ? itr->second : static_cast<coefficient_t>(0);
	}

	// Number to scalar multivector conversion.
	inline multivector scalar(coefficient_t arg) {
		return multivector { { {mask_t(0), arg} } };
	}

	// Return the i-th basis vector.
	inline multivector e(index_t index) {
		return multivector { { {mask_t(1) << index, static_cast<coefficient_t>(1)} } };
	}

	namespace detail {

		// Population (number of 1 bits) count.
		inline int _population_count(uint64_t w) {
			uint64_t w1 = (w & 0x2222222222222222) + ((w + w) & 0x2222222222222222);
			uint64_t w2 = (w >> 1 & 0x2222222222222222) + (w >> 2 & 0x2222222222222222);
			w1 += (w1 >> 4) & 0x0f0f0f0f0f0f0f0f;
			w2 += (w2 >> 4) & 0x0f0f0f0f0f0f0f0f;
			return (int)((w1 + w2) * 0x0101010101010101 >> 57);
		}

		// General implementation of bilinear products.
		inline multivector _graded_gp(const multivector& lhs, const multivector& rhs, const orthogonal_metric& metric, std::function<bool(grade_t, grade_t, grade_t)> keep) {
			auto reordering_sign = [&] (mask_t lhs_mask, mask_t rhs_mask) {
					lhs_mask >>= 1;
					int changes = 0;
					while (lhs_mask != mask_t(0)) {
						changes += detail::_population_count(lhs_mask & rhs_mask);
						lhs_mask >>= 1;
					}
					return static_cast<entry_t>((changes & 1) == 0 ? +1 : -1);
				};

			multivector result;
			for (auto& lhs_comp : lhs.components_) {
				for (auto& rhs_comp : rhs.components_) {
					mask_t mask = lhs_comp.first ^ rhs_comp.first;
					if (keep(lhs_comp.first, rhs_comp.first, mask)) {
						coefficient_t coef = reordering_sign(lhs_comp.first, rhs_comp.first) * metric.metric_factor(lhs_comp.first & rhs_comp.first) * lhs_comp.second * rhs_comp.second;
						auto itr = result.components_.find(mask);
						if (itr != result.components_.end()) {
							itr->second += coef;
						}
						else {
							result.components_.emplace(mask, coef);
						}
					}
				}
			}
			return result;
		}

		inline multivector _graded_gp(coefficient_t lhs, const multivector& rhs, const orthogonal_metric& metric, std::function<bool(grade_t, grade_t, grade_t)> keep) {
			return _graded_gp(scalar(lhs), rhs, metric, keep);
		}

		inline multivector _graded_gp(const multivector& lhs, coefficient_t rhs, const orthogonal_metric& metric, std::function<bool(grade_t, grade_t, grade_t)> keep) {
			return _graded_gp(lhs, scalar(rhs), metric, keep);
		}

		inline multivector _graded_gp(coefficient_t lhs, coefficient_t rhs, const orthogonal_metric& metric, std::function<bool(grade_t, grade_t, grade_t)> keep) {
			return _graded_gp(scalar(lhs), scalar(rhs), metric, keep);
		}

		// General implementation of grade-based sign-change operations.
		inline multivector _graded_uminus(const multivector& rhs, std::function<bool(grade_t)> change) {
			multivector result;

			for (auto itr = rhs.components_.begin(), end = rhs.components_.end(); itr != end; ++itr) {
				if (change(itr->first)) {
					result.components_.emplace(itr->first, -itr->second);
				}
				else {
					result.components_.emplace(itr->first, itr->second);
				}
			}

			return result;
		}

		inline multivector _graded_uminus(coefficient_t rhs, std::function<bool(grade_t)> change) {
			return _graded_uminus(scalar(rhs), change);
		}

		// General implementation of plus and minus operations.
		inline multivector _merge(const multivector& lhs, const multivector& rhs, std::function<coefficient_t(coefficient_t, coefficient_t)> func) {
			multivector result;

			auto lhs_itr = lhs.components_.begin();
			auto rhs_itr = rhs.components_.begin();
			while (lhs_itr != lhs.components_.end() && rhs_itr != rhs.components_.end()) {
				if (lhs_itr->first < rhs_itr->first) {
					result.components_.emplace(lhs_itr->first, func(lhs_itr->second, static_cast<coefficient_t>(0)));
					++lhs_itr;
				}
				else if (lhs_itr->first > rhs_itr->first) {
					result.components_.emplace(rhs_itr->first, func(static_cast<coefficient_t>(0), rhs_itr->second));
					++rhs_itr;
				}
				else {
					result.components_.emplace(lhs_itr->first, func(lhs_itr->second, rhs_itr->second));
					++lhs_itr;
					++rhs_itr;
				}
			}

			for (auto end = lhs.components_.end(); lhs_itr != end; ++lhs_itr) {
				result.components_.emplace(lhs_itr->first, func(lhs_itr->second, static_cast<coefficient_t>(0)));
			}

			for (auto end = rhs.components_.end(); rhs_itr != end; ++rhs_itr) {
				result.components_.emplace(rhs_itr->first, func(static_cast<coefficient_t>(0), rhs_itr->second));
			}

			return result;
		}

		inline multivector _merge(coefficient_t lhs, const multivector& rhs, std::function<coefficient_t(coefficient_t, coefficient_t)> func) {
			return _merge(scalar(lhs), rhs, func);
		}

		inline multivector _merge(const multivector& lhs, coefficient_t rhs, std::function<coefficient_t(coefficient_t, coefficient_t)> func) {
			return _merge(lhs, scalar(rhs), func);
		}

		inline multivector _merge(coefficient_t lhs, coefficient_t rhs, std::function<coefficient_t(coefficient_t, coefficient_t)> func) {
			return _merge(scalar(lhs), scalar(rhs), func);
		}

		// Extract the basis vectors used by the given mutivector.
		inline mask_t _used_basis_vectors(const multivector& arg) {
		    mask_t bits = 0;
		    for (auto itr = arg.begin(), end = arg.end(); itr != end; ++itr) {
		        bits ^= itr->first;
		    }
		    return bits;
		}

	}

	// Equality check function.
	inline bool equal(const multivector& lhs, const multivector& rhs, coefficient_t tol = default_tolerance()) {
		auto lhs_itr = lhs.begin();
		auto rhs_itr = rhs.begin();
		while (lhs_itr != lhs.end() && rhs_itr != rhs.end()) {
			if (lhs_itr->first < rhs_itr->first) {
				if (abs(lhs_itr->second) > tol) {
					return false;
				}
				++lhs_itr;
			}
			else if (lhs_itr->first > rhs_itr->first) {
				if (abs(rhs_itr->second) > tol) {
					return false;
				}
				++rhs_itr;
			}
			else {
				if (abs(lhs_itr->second - rhs_itr->second) > tol) {
					return false;
				}
				++lhs_itr;
				++rhs_itr;
			}
		}

		for (auto end = lhs.end(); lhs_itr != end; ++lhs_itr) {
			if (abs(lhs_itr->second) > tol) {
				return false;
			}
		}

		for (auto end = rhs.end(); rhs_itr != end; ++rhs_itr) {
			if (abs(rhs_itr->second) > tol) {
				return false;
			}
		}

		return true;
	}

	inline bool equal(coefficient_t lhs, const multivector& rhs, coefficient_t tol = default_tolerance()) {
		return equal(scalar(lhs), rhs, tol);
	}

	inline bool equal(const multivector& lhs, coefficient_t rhs, coefficient_t tol = default_tolerance()) {
		return equal(lhs, scalar(rhs), tol);
	}

	// Return the grade of the multivector, -1 for mixed grade, or -2 for zero multivector.
	inline grade_t grade(const multivector& arg, coefficient_t tol = default_tolerance()) {
		grade_t result = -2;
		for (auto itr = arg.begin(); itr != arg.end(); ++itr) {
	        if (abs(itr->second) > tol) {
	        	grade_t curr = detail::_population_count(itr->first);
	            if (result != -2) {
	                if (result != curr) {
	                    return -1;
	                }
	            }
	            else {
	            	result = curr;
	            }
	        }
		}
	    return result;
	}

	inline grade_t grade(coefficient_t arg) {
		return grade(scalar(arg));
	}

	// Grade extraction operation.
	inline multivector take_grade(const multivector& arg, grade_t k) {
		multivector result;
		std::copy_if(arg.components_.begin(), arg.components_.end(), std::inserter(result.components_, result.components_.end()), [&] (const std::pair<mask_t, coefficient_t>& comp) {
				return detail::_population_count(comp.first) == k;
			});
		return result;
	}

	inline multivector take_grade(coefficient_t arg, grade_t k) {
		return take_grade(scalar(arg), k);
	}

	// Geometric product.
	template<class LeftType, class RightType, class MetricType>
	inline multivector gp(LeftType&& lhs, RightType&& rhs, MetricType&& metric) {
		return detail::_graded_gp(lhs, rhs, metric, [] (grade_t, grade_t, grade_t) -> bool {
					return true;
				}
			);
	}

	// Geometric product under Euclidean metric.
	template<class LeftType, class RightType>
	inline multivector gp_em(LeftType&& lhs, RightType&& rhs) {
		return gp(lhs, rhs, euclidean_metric());
	}

	// Geometric product where one of the arguments is a scalar value.
	inline multivector operator* (const multivector& lhs, const multivector& rhs) {
		if (!is_scalar(lhs) && !is_scalar(rhs)) {
			throw std::invalid_argument("One of the arguments must be a scalar value.");
		}
		return gp_em(lhs, rhs);
	}

	inline multivector operator* (coefficient_t lhs, const multivector& rhs) {
		return gp_em(lhs, rhs);
	}

	inline multivector operator* (const multivector& lhs, coefficient_t rhs) {
		return gp_em(lhs, rhs);
	}

	// Outer product.
	template<class LeftType, class RightType>
	inline multivector op(LeftType&& lhs, RightType&& rhs) {
		return detail::_graded_gp(lhs, rhs, euclidean_metric(), [] (grade_t lhs_grade, grade_t rhs_grade, grade_t result_grade) -> bool {
					return result_grade == (lhs_grade + rhs_grade);
				}
			);
	}

	// Outer product.
	inline multivector operator^ (const multivector& lhs, const multivector& rhs) {
		return op(lhs, rhs);
	}

	inline multivector operator^ (coefficient_t lhs, const multivector& rhs) {
		return op(lhs, rhs);
	}

	inline multivector operator^ (const multivector& lhs, coefficient_t rhs) {
		return op(lhs, rhs);
	}

	// Left contraction.
	template<class LeftType, class RightType, class MetricType>
	inline multivector lcont(LeftType&& lhs, RightType&& rhs, MetricType&& metric) {
		return detail::_graded_gp(lhs, rhs, metric, [] (grade_t lhs_grade, grade_t rhs_grade, grade_t result_grade) -> bool {
					return result_grade == (rhs_grade - lhs_grade);
				}
			);
	}

	// Left contraction under Euclidean metric.
	template<class LeftType, class RightType>
	inline multivector lcont_em(LeftType&& lhs, RightType&& rhs) {
		return lcont(lhs, rhs, euclidean_metric());
	}

	// Right contraction.
	template<class LeftType, class RightType, class MetricType>
	inline multivector rcont(LeftType&& lhs, RightType&& rhs, MetricType&& metric) {
		return detail::_graded_gp(lhs, rhs, metric, [] (grade_t lhs_grade, grade_t rhs_grade, grade_t result_grade) -> bool {
					return result_grade == (lhs_grade - rhs_grade);
				}
			);
	}

	// Right contraction under Euclidean metric.
	template<class LeftType, class RightType>
	inline multivector rcont_em(LeftType&& lhs, RightType&& rhs) {
		return rcont(lhs, rhs, euclidean_metric());
	}

	// Scalar product.
	template<class LeftType, class RightType, class MetricType>
	inline multivector scp(LeftType&& lhs, RightType&& rhs, MetricType&& metric) {
		return detail::_graded_gp(lhs, rhs, metric, [] (grade_t, grade_t, grade_t result_grade) -> bool {
					return result_grade == 0;
				}
			);
	}

	// Scalar product under Euclidean metric.
	template<class LeftType, class RightType>
	inline multivector scp_em(LeftType&& lhs, RightType&& rhs) {
		return scp(lhs, rhs, euclidean_metric());
	}

	// Delta product.
	template<class LeftType, class RightType, class MetricType>
	inline multivector delta(LeftType&& lhs, RightType&& rhs, MetricType&& metric, coefficient_t tol = default_tolerance()) {
		auto m = gp(lhs, rhs, metric);

		int max_grade = -1;
		for (auto itr = m.begin(), end = m.end(); itr != end; ++itr) {
			if (abs(itr->second) > tol) {
				int curr_grade = detail::_population_count(itr->first);
				if (max_grade < curr_grade) {
					max_grade = curr_grade;
				}
			}
		}

		return take_grade(m, max_grade);
	}

	// Delta product under Euclidean metric.
	template<class LeftType, class RightType>
	inline multivector delta_em(LeftType&& lhs, RightType&& rhs, coefficient_t tol = default_tolerance()) {
		return delta(lhs, rhs, euclidean_metric(), tol);
	}

	// Unary plus.
	inline multivector uplus(const multivector& rhs) {
		return rhs;
	}

	// Unary plus.
	inline multivector operator+ (const multivector& rhs) {
		return uplus(rhs);
	}

	// Plus.
	template<class LeftType, class RightType>
	inline multivector plus(LeftType&& lhs, RightType&& rhs) {
		return detail::_merge(lhs, rhs, [] (coefficient_t lhs_coef, coefficient_t rhs_coef) -> coefficient_t {
					return lhs_coef + rhs_coef;
				}
			);
	}

	// Plus.
	inline multivector operator+ (const multivector& lhs, const multivector& rhs) {
		return plus(lhs, rhs);
	}

	inline multivector operator+ (coefficient_t lhs, const multivector& rhs) {
		return plus(lhs, rhs);
	}

	inline multivector operator+ (const multivector& lhs, coefficient_t rhs) {
		return plus(lhs, rhs);
	}

	// Unary minus.
	inline multivector uminus(const multivector& rhs) {
		return detail::_graded_uminus(rhs, [] (grade_t) -> bool {
					return true;
				}
			);
	}

	// Unary minus.
	inline multivector operator- (const multivector& rhs) {
		return uminus(rhs);
	}

	// Minus.
	template<class LeftType, class RightType>
	inline multivector minus(LeftType&& lhs, RightType&& rhs) {
		return detail::_merge(lhs, rhs, [] (coefficient_t lhs_coef, coefficient_t rhs_coef) -> coefficient_t {
					return lhs_coef - rhs_coef;
				}
			);
	}

	// Minus.
	inline multivector operator- (const multivector& lhs, const multivector& rhs) {
		return minus(lhs, rhs);
	}

	inline multivector operator- (coefficient_t lhs, const multivector& rhs) {
		return minus(lhs, rhs);
	}

	inline multivector operator- (const multivector& lhs, coefficient_t rhs) {
		return minus(lhs, rhs);
	}

	// Clifford conjugation operation.
	inline multivector conjugation(const multivector& arg) {
		return detail::_graded_uminus(arg, [] (mask_t mask) -> bool {
					int k = detail::_population_count(mask);
					return (((k * (k + 1)) >> 1) & 1) != 0;
				}
			);
	}

	inline multivector conjugation(coefficient_t arg) {
		return scalar(arg);
	}

	// Grade involution operation.
	inline multivector involution(const multivector& arg) {
		return detail::_graded_uminus(arg, [] (mask_t mask) -> bool {
					return (detail::_population_count(mask) & 1) != 0;
				}
			);
	}

	inline multivector involution(coefficient_t arg) {
		return scalar(arg);
	}

	// Reversion operation.
	inline multivector reversion(const multivector& arg) {
		return detail::_graded_uminus(arg, [] (mask_t mask) -> bool {
					int k = detail::_population_count(mask);
					return (((k * (k - 1)) >> 1) & 1) != 0;
				}
			);
	}

	inline multivector reversion(coefficient_t arg) {
		return scalar(arg);
	}

	// Squared reverse norm.
	template<class Type, class MetricType>
	inline coefficient_t sqr_rnorm(Type&& arg, MetricType&& metric, coefficient_t tol = default_tolerance()) {
		return native(scp(arg, reversion(arg), metric), tol);
	}

	// Squared reverse norm under Euclidean metric.
	template<class Type>
	inline coefficient_t sqr_rnorm_em(Type&& arg, coefficient_t tol = default_tolerance()) {
		return sqr_rnorm(arg, euclidean_metric(), tol);
	}

	// Reverse norm.
	template<class Type, class MetricType>
	inline coefficient_t rnorm(Type&& arg, MetricType&& metric, coefficient_t tol = default_tolerance()) {
		return sqrt(sqr_rnorm(arg, metric, tol));
	}

	// Reverse norm under Euclidean metric.
	template<class Type>
	inline coefficient_t rnorm_em(Type&& arg, coefficient_t tol = default_tolerance()) {
		return rnorm(arg, euclidean_metric(), tol);
	}

	// Blade inverse.
	template<class Type, class MetricType>
	inline multivector inv(Type&& arg, MetricType&& metric, coefficient_t tol = default_tolerance()) {
	    auto rev = reversion(arg);
	    return gp(rev, static_cast<coefficient_t>(1) / native(scp(arg, rev, metric), tol), metric);
	}

    // Blade inverse under Euclidean metric.
	template<class Type>
	inline multivector inv_em(Type&& arg, coefficient_t tol = default_tolerance()) {
	    return inv(arg, euclidean_metric(), tol);
	}

	// Inverse geometric product.
	template<class LeftType, class RightType, class MetricType>
	inline multivector igp(LeftType&& lhs, RightType&& rhs, MetricType&& metric, coefficient_t tol = default_tolerance()) {
		return gp(lhs, inv(rhs, metric, tol), metric);
	}

	// Inverse geometric product under Euclidean metric.
	template<class LeftType, class RightType>
	inline multivector igp_em(LeftType&& lhs, RightType&& rhs, coefficient_t tol = default_tolerance()) {
		return igp(lhs, rhs, euclidean_metric(), tol);
	}

	// Dualization.
	template<class Type, class PseudoscalarType, class MetricType>
	inline multivector dual(Type&& arg, PseudoscalarType&& pseudoscalar, MetricType&& metric) {
		return gp(arg, inv(pseudoscalar, metric), metric);
	}

	// Dualization under Euclidean metric.
	template<class Type, class PseudoscalarType>
	inline multivector dual_em(Type&& arg, PseudoscalarType&& pseudoscalar) {
		return dual(arg, pseudoscalar, euclidean_metric());
	}

	// Undualization.
	template<class Type, class PseudoscalarType, class MetricType>
	inline multivector undual(Type&& arg, PseudoscalarType&& pseudoscalar, MetricType&& metric) {
		return gp(arg, pseudoscalar, metric);
	}

	// Undualization under Euclidean metric.
	template<class Type, class PseudoscalarType>
	inline multivector undual_em(Type&& arg, PseudoscalarType&& pseudoscalar) {
		return undual(arg, pseudoscalar, euclidean_metric());
	}

	// Returns whether the given multivetor is a versor.
	inline bool is_versor(const multivector& arg, const orthogonal_metric& metric, coefficient_t tol = default_tolerance()) {
		auto arg_hat = involution(arg);
		auto inv_arg = inv(arg, metric, tol);

		auto tmp1 = gp(arg_hat, inv_arg, metric);
		if (grade(tmp1, tol) != 0) {
			return false;
		}

		auto tmp2 = gp(inv_arg, arg_hat, metric);
		if (!equal(tmp1, tmp2, tol)) {
			return false;
		}

		auto arg_tild = reversion(arg);

		index_t index = 0;
		mask_t basis_vectors = detail::_used_basis_vectors(arg);

		while (basis_vectors != mask_t(0)) {
			if ((basis_vectors & mask_t(1)) != mask_t(0) && grade(gp(gp(arg_hat, e(index), metric), arg_tild, metric), tol) != 1) {
				return false;
			}
			basis_vectors >>= 1;
			++index;
		}

		return true;
	}

	inline bool is_versor(coefficient_t arg, const orthogonal_metric& metric, coefficient_t tol = default_tolerance()) {
		return is_versor(scalar(arg), metric, tol);
	}

	// Returns whether the given multivetor is a versor under Euclidean metric.
	template<class Type>
	inline bool is_versor_em(Type&& arg, coefficient_t tol = default_tolerance()) {
		return is_versor(arg, euclidean_metric(), tol);
	}

	// Returns whether the given multivetor is a blade.
	template<class Type>
	inline bool is_blade(Type&& arg, coefficient_t tol = default_tolerance()) {
		return grade(arg, tol) != -1 && is_versor_em(arg, tol);
	}

	// Write operation.
	inline std::ostream& operator<< (std::ostream& os, const multivector& rhs) {
		bool is_first_comp = true;

		for (auto& comp : rhs.components_) {
			coefficient_t abs_coef = std::abs(comp.second);
			if (abs_coef != coefficient_t(0)) {
				bool is_negative = (comp.second < static_cast<coefficient_t>(0));
				bool is_one = (abs_coef == static_cast<coefficient_t>(1));
				bool is_scalar = (comp.first == mask_t(0));
	
				if (!is_first_comp) {
					if (is_negative) os << " - ";
					else os << " + ";
				}
				else {
					if (is_negative) os << "-";
					is_first_comp = false;
				}
	
				if (!is_one || is_scalar) os << abs_coef;
				if (!is_one && !is_scalar) os << " * ";
	
				if (!is_scalar) {
					bool is_first_factor = true;
					index_t index = 0;
					mask_t mask = comp.first;
					while (mask != mask_t(0)) {
						if ((mask & mask_t(1)) != mask_t(0)) {
							if (!is_first_factor) os << "^";
							os << "e" << index;
							is_first_factor = false;
						}
						index++;
						mask >>= 1;
					}
	
				}
			}
		}
		
		if (is_first_comp) {
			os << static_cast<coefficient_t>(0);
		}

	    return os;
	}

}

#endif // GA_HPP
